import datetime
from .channel import Channel, Message
from .emoji import Emoji
from .gateway import Activity
from .guild import Guild, GuildMember
from .permission import Role
from .snowflake import Snowflake
from .user import User
from .voice import VoiceState
from .interactions import Interaction
from ..base.model import EventBase


class Ready(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.v = resp["v"]
        self.user = User.create(client, resp["user"])
        self.guilds = resp["guilds"]
        self.session_id = resp["session_id"]
        self.shard = resp.get("shard", [])
        self.application = resp["application"]


class ApplicationCommandCreate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)


class ApplicationCommandUpdate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)


class ApplicationCommandDelete(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)


ChannelCreate = Channel


class ChannelUpdate(Channel):
    def __del__(self):
        Channel.create(self.client, self.raw)

    @classmethod
    def create(cls, client, resp, **kwargs):
        return cls(client, resp, **kwargs)

    @property
    def original(self):
        if self.client.has_cache:
            return self.client.get_channel(self.id)


class ChannelDelete(Channel):
    def __del__(self):
        if self.client.has_cache:
            self.client.cache.remove(self.id, self._cache_type)


class ChannelPinsUpdate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake.optional(resp.get("guild_id"))
        self.channel_id = Snowflake(resp["channel_id"])
        self.__last_pin_timestamp = resp.get("last_pin_timestamp")
        self.last_pin_timestamp = datetime.datetime.fromisoformat(self.__last_pin_timestamp) if self.__last_pin_timestamp else self.__last_pin_timestamp

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def guild(self):
        if self.guild_id:
            return self.client.get(self.guild_id, "guild")


GuildCreate = Guild
GuildUpdate = Guild
GuildDelete = Guild


class GuildBanAdd(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])
        self.user = User.create(self.client, resp["user"])

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class GuildBanRemove(GuildBanAdd):
    pass


class GuildEmojisUpdate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])
        self.emojis = [Emoji(self.client, x) for x in resp["emojis"]]

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class GuildIntegrationsUpdate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


GuildMemberAdd = GuildMember


class GuildMemberRemove(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])
        self.user = User.create(self.client, resp["user"])

    def __del__(self):
        if self.client.has_cache:
            self.guild.cache.remove(self.user.id, "member")

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")

    @property
    def member(self):
        if self.client.has_cache:
            return self.guild.get(self.user.id, "member")


class GuildMemberUpdate(GuildMember):
    def __del__(self):
        super().create(self.client, self.raw, user=self.user, guild_id=self.guild_id, cache=True)

    @classmethod
    def create(cls, client, resp, *, user=None, guild_id=None, cache: bool = False):
        return super().create(client, resp, user=user, guild_id=guild_id, cache=False)

    @property
    def original(self):
        if self.client.has_cache:
            return self.client.cache.get_guild_container(self.guild_id).get_storage("member").get(self.user.id)


class GuildRoleCreate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])
        self.role = Role.create(client, resp["role"], guild_id=self.guild_id)

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class GuildRoleUpdate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])
        self.role = Role(client, resp["role"], guild_id=self.guild_id)

    def __del__(self):
        Role.create(self.client, self.raw["role"], guild_id=self.guild_id)

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")

    @property
    def original(self):
        if self.client.has_cache:
            return self.guild.get(self.role.id, "role")


class GuildRoleDelete(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.guild_id = Snowflake(resp["guild_id"])
        self.role_id = Snowflake(resp["role_id"])

    def __del__(self):
        if self.client.has_cache:
            self.client.cache.remove(self.role_id, "role")

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")

    @property
    def role(self):
        if self.client.has_cache:
            return self.guild.get(self.role_id, "role")


InteractionCreate = Interaction


class InviteCreate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.channel_id = Snowflake(resp["channel_id"])
        self.code = resp["code"]
        self.created_at = datetime.datetime.fromisoformat(resp["created_at"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))
        self.__inviter = resp.get("inviter")
        self.inviter = User.create(client, self.__inviter) if self.__inviter else self.__inviter
        self.max_age = resp["max_age"]
        self.max_uses = resp["max_uses"]
        self.target_type = resp.get("target_type")
        self.__target_user = resp.get("target_user")
        self.target_user = User.create(client, self.__target_user) if self.__target_user else self.__target_user
        self.target_application = resp.get("target_application")
        self.temporary = resp["temporary"]
        self.uses = resp["uses"]

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class InviteDelete(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.channel_id = Snowflake(resp["channel_id"])
        self.code = resp["code"]
        self.guild_id = Snowflake.optional(resp.get("guild_id"))

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


MessageCreate = Message


class MessageUpdate(Message):
    def __del__(self):
        Message.create(self.client, self.raw, guild_id=self.guild_id)

    @classmethod
    def create(cls, client, resp, **kwargs):
        try:
            return cls(client, resp, **kwargs)
        except KeyError:
            if client.has_cache:
                msg = client.get(resp["id"])
                if msg:
                    orig = msg.raw
                    for k, v in resp.items():
                        if orig.get(k) != v:
                            orig[k] = v
                    ret = cls(client, orig, **kwargs)
                    ret._dont_dispatch = True
                    return ret

    @property
    def original(self):
        if self.client.has_cache:
            return self.client.get(self.id, "message")


class MessageDelete(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.id = Snowflake(resp["id"])
        self.channel_id = Snowflake(resp["channel_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))

    def __del__(self):
        if self.client.has_cache:
            self.client.cache.remove(self.id, "message")

    @property
    def message(self):
        if self.client.has_cache:
            return self.client.get(self.id, "message")

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class MessageDeleteBulk(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.ids = [Snowflake(x) for x in resp["ids"]]
        self.channel_id = Snowflake(resp["channel_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))

    def __del__(self):
        if self.client.has_cache:
            [self.client.cache.remove(x.id, "message") for x in self.available_messages]

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")

    @property
    def available_messages(self):
        if self.client.has_cache:
            tries = [self.client.get(x, "message") for x in self.ids]
            return [x for x in tries if x]


class MessageReactionAdd(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.user_id = Snowflake(resp["user_id"])
        self.channel_id = Snowflake(resp["channel_id"])
        self.message_id = Snowflake(resp["message_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))
        self.__member = resp.get("member")
        self.member = GuildMember.create(client, self.__member, guild_id=self.guild_id)
        self.emoji = Emoji(client, resp["emoji"])

    @property
    def user(self):
        if self.client.has_cache:
            return self.client.get(self.user_id, "user")

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def message(self):
        if self.client.has_cache:
            return self.client.get(self.message_id, "message")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class MessageReactionRemove(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.user_id = Snowflake(resp["user_id"])
        self.channel_id = Snowflake(resp["channel_id"])
        self.message_id = Snowflake(resp["message_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))
        self.emoji = Emoji(client, resp["emoji"])

    @property
    def user(self):
        if self.client.has_cache:
            return self.client.get(self.user_id, "user")

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def message(self):
        if self.client.has_cache:
            return self.client.get(self.message_id, "message")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class MessageReactionRemoveAll(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.channel_id = Snowflake(resp["channel_id"])
        self.message_id = Snowflake(resp["message_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def message(self):
        if self.client.has_cache:
            return self.client.get(self.message_id, "message")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class MessageReactionRemoveEmoji(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.channel_id = Snowflake(resp["channel_id"])
        self.message_id = Snowflake(resp["message_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))
        self.emoji = Emoji(client, resp["emoji"])

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def message(self):
        if self.client.has_cache:
            return self.client.get(self.message_id, "message")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class PresenceUpdate(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.user = User.create(self.client, resp["user"])
        self.guild_id = Snowflake(resp["guild_id"])
        self.status = resp["status"]
        self.activities = [Activity(x) for x in resp["activities"]]
        self.client_status = resp["client_status"]

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.cache.get(self.guild_id, "guild")


class TypingStart(EventBase):
    def __init__(self, client, resp: dict):
        super().__init__(client, resp)
        self.channel_id = Snowflake(resp["channel_id"])
        self.guild_id = Snowflake.optional(resp.get("guild_id"))
        self.user_id = Snowflake(resp["user_id"])
        self.timestamp = datetime.datetime.fromtimestamp(resp["timestamp"])

    @property
    def channel(self):
        if self.client.has_cache:
            return self.client.get(self.channel_id, "channel")

    @property
    def guild(self):
        if self.guild_id and self.client.has_cache:
            return self.client.get(self.guild_id, "guild")

    @property
    def user(self):
        if self.client.has_cache:
            return self.client.get(self.user_id, "user")


class UserUpdate(User):
    def __del__(self):
        User.create(self.client, self.raw)

    @classmethod
    def create(cls, client, resp, **kwargs):
        return cls(client, resp)

    @property
    def original(self):
        if self.client.has_cache:
            return self.client.get(self.id, "guild")


VoiceStateUpdate = VoiceState



import time
import typing
import asyncio
import logging
import aiohttp
from ..http.async_http import AsyncHTTPRequest
from ..model import gateway
from ..handler import EventHandler


class WebSocketClient:
    def __init__(self,
                 http: AsyncHTTPRequest,
                 ws: aiohttp.ClientWebSocketResponse,
                 base_url: str,
                 intents: typing.Union[gateway.Intents, int],
                 event_handler: EventHandler,
                 try_reconnect: bool):
        self.ws = ws
        self.http = http
        self.base_url = base_url
        self.event_handler = event_handler
        self.logger = logging.getLogger("dico.ws")
        self._closed = False
        self._reconnecting = False
        self._fresh_reconnecting = False
        self.heartbeat_interval = 0
        self.seq = None
        self.last_heartbeat_ack = 0
        self.last_heartbeat_send = 0
        self._heartbeat_task = None
        self.session_id = None
        self.intents = intents
        self.ping = 0.0
        self._ping_start = 0.0
        self.try_reconnect = try_reconnect

    def __del__(self):
        if not self._closed:
            self.logger.warning("Please properly close before exiting!")
            self.http.loop.run_until_complete(self.close())

    async def close(self, code: int = 1000):
        if not self._heartbeat_task.cancelled():
            self._heartbeat_task.cancel()
        if self._closed:
            return
        await self.ws.close(code=code)
        self._closed = True

    async def run(self):
        while True:
            while not self._closed:
                try:
                    resp = await self.receive()
                except WSClosing as ex:
                    self.logger.warning(f"Websocket is closing with code: {ex.code}")
                    if ex.code is None and self.try_reconnect:
                        self.logger.warning("Trying to reconnect...")
                        await self.reconnect(fresh=True)
                    break
                self.logger.debug(f"Received `{gateway.Opcodes.as_string(resp.op)}` payload"
                                  f"{f' with event name `{resp.t}`' if resp.op == gateway.Opcodes.DISPATCH else ''}.")
                if resp.s:
                    self.seq = resp.s

                if resp.op == gateway.Opcodes.DISPATCH:
                    if resp.t == "READY":
                        self.session_id = resp.d.get("session_id", self.session_id)
                    self.event_handler.client.dispatch("RAW", resp.raw)
                    self.event_handler.dispatch_from_raw(resp.t, resp.d)

                elif resp.op == gateway.Opcodes.HELLO:
                    self.heartbeat_interval = resp.d["heartbeat_interval"]
                    self._heartbeat_task = self.http.loop.create_task(self.run_heartbeat())
                    if self._reconnecting:
                        await self.resume()
                    else:
                        await self.identify()
                    if self._fresh_reconnecting:
                        self._fresh_reconnecting = False

                elif resp.op == gateway.Opcodes.HEARTBEAT_ACK:
                    self.last_heartbeat_ack = time.time()
                    self.ping = self.last_heartbeat_ack - self._ping_start

                elif resp.op == gateway.Opcodes.INVALID_SESSION:
                    self.logger.warning("Failed gateway resume, reconnecting to gateway without resuming.")
                    await asyncio.sleep(5)
                    await self.reconnect(fresh=True)
                    break

                elif resp.op == gateway.Opcodes.RECONNECT:
                    await self.reconnect()
                    break

            if self._reconnecting or self._fresh_reconnecting:
                self.ws = await self.http.session.ws_connect(self.base_url)
                self._closed = False
            else:
                return

    async def receive(self):
        resp = await self.ws.receive()
        if resp.type == aiohttp.WSMsgType.TEXT:
            res = gateway.GatewayResponse(resp.json())
            self.seq = res.s
            return res
        elif resp.type in (aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSED):  # Somehow this is not mentioned in aiohttp document hmm.
            raise WSClosing(resp.data)

    async def reconnect(self, fresh: bool = False):
        if self._reconnecting or self._fresh_reconnecting:
            self.logger.warning("Reconnection is already running, but another reconnection is requested. This request is ignored.")
            return
        self._reconnecting = not fresh
        self._fresh_reconnecting = fresh
        self.logger.info("Reconnecting to Websocket...")
        if not self._closed:
            await self.close(4000)

    async def resume(self):
        data = {
            "op": gateway.Opcodes.RESUME,
            "d": {
                "token": self.http.token,
                "session_id": self.session_id,
                "seq": self.seq
            }
        }
        await self.ws.send_json(data)
        self._reconnecting = False

    async def run_heartbeat(self):
        while not self._closed:
            if not self.last_heartbeat_send <= self.last_heartbeat_ack <= time.time():
                await self.reconnect()
                break
            data = {"op": gateway.Opcodes.HEARTBEAT, "d": self.seq}
            self._ping_start = time.time()
            await self.ws.send_json(data)
            self.last_heartbeat_send = time.time()
            await asyncio.sleep(self.heartbeat_interval/1000)

    async def identify(self):
        data = {
            "op": gateway.Opcodes.IDENTIFY,
            "d": {
                "token": self.http.token,
                "intents": int(self.intents),
                "properties": {
                    "$os": "linux",
                    "$browser": "dico",
                    "$device": "dico"
                }
            }
        }
        await self.ws.send_json(data)

    async def request_guild_members(self, guild_id: str, *, query: str = None, limit: int = None, presences: bool = None, user_ids: typing.List[str] = None, nonce: bool = None):
        d = {"guild_id": guild_id}
        if query is not None:
            d["query"] = query
            d["limit"] = limit
        if presences is not None:
            d["presences"] = presences
        if user_ids is not None:
            d["user_ids"] = user_ids
        if nonce is not None:
            d["nonce"] = nonce
        data = {
            "op": gateway.Opcodes.REQUEST_GUILD_MEMBERS,
            "d": d
        }
        await self.ws.send_json(data)

    async def update_voice_state(self, guild_id: str, channel_id: typing.Optional[str], self_mute: bool, self_deaf: bool):
        data = {
            "op": gateway.Opcodes.VOICE_STATE_UPDATE,
            "d": {
                "guild_id": guild_id,
                "channel_id": channel_id,
                "self_mute": self_mute,
                "self_deaf": self_deaf
            }
        }
        await self.ws.send_json(data)

    async def update_presence(self, since: typing.Optional[int], activities: typing.List[dict], status: str, afk: bool):
        data = {
            "op": gateway.Opcodes.PRESENCE_UPDATE,
            "d": {
                "since": since,
                "activities": activities,
                "status": status,
                "afk": afk
            }
        }
        await self.ws.send_json(data)

    @property
    def closed(self):
        return self._closed

    @classmethod
    async def connect(cls, http, intents, event_handler, try_reconnect):
        resp = await http.request("/gateway/bot", "GET")
        gw = gateway.GetGateway(resp)
        base_url = gw.url+f"?v=9&encoding=json"
        ws = await http.session.ws_connect(base_url)
        return cls(http, ws, base_url, intents, event_handler, try_reconnect)


class WSClosing(Exception):
    """Internal exception class for controlling WS close event. You should not see this exception in normal situation."""
    def __init__(self, code):
        self.code = code



import datetime


class Snowflake:
    def __init__(self, snowflake):
        self.__snowflake = int(snowflake)

    @property
    def timestamp(self):
        return datetime.datetime.fromtimestamp(((self.__snowflake >> 22) + 1420070400000)/1000)

    @property
    def increment(self):
        return self.__snowflake & 0xFFF

    @property
    def wid(self):
        return (self.__snowflake & 0x3E0000) >> 17

    @property
    def pid(self):
        return (self.__snowflake & 0x1F000) >> 12

    @property
    def id(self):
        return self.__snowflake

    def __str__(self):
        return str(self.__snowflake)

    def __int__(self):
        return self.__snowflake

    def __eq__(self, other):
        return self.__snowflake == int(other)

    def __ne__(self, other):
        return self.__snowflake != int(other)

    def __lt__(self, other):
        return self.__snowflake < int(other)

    def __le__(self, other):
        return self.__snowflake <= int(other)

    def __gt__(self, other):
        return self.__snowflake > int(other)

    def __ge__(self, other):
        return self.__snowflake >= int(other)

    def __hash__(self):
        return hash(self.__snowflake)

    @classmethod
    def optional(cls, snowflake):
        return cls(snowflake) if snowflake else snowflake

    @classmethod
    def ensure_snowflake(cls, target):
        return target if isinstance(target, cls) else cls.optional(target)



from .channel import *
from .emoji import *
from .event import *
from .gateway import *
from .guild import *
from .invite import *
from .permission import *
from .snowflake import *
from .user import *
from .webhook import *
from .interactions import *



from .utils import format_discord_error


class DicoException(Exception):
    """Base exception class for this library."""


class WebsocketClosed(DicoException):
    """Websocket is closed, so this action could not be performed."""


class HTTPError(DicoException):
    """Special exception class for HTTP."""
    def __init__(self, route, code, resp):
        self.route = route
        self.code = code
        self.resp = resp
        super().__init__(format_discord_error(self.resp))


class DiscordError(HTTPError):
    """Discord has an error."""


class BadRequest(HTTPError):
    """We sent incorrect request."""


class Forbidden(HTTPError):
    """We don't have permission for this request."""


class NotFound(HTTPError):
    """This request isn't something available."""


class RateLimited(HTTPError):
    """We are rate-limited and couldn't successfully requesting in desired count."""


class Unknown(HTTPError):
    """Failed handling this type of error. If you see this report to GitHub Issues."""



from .user import User
from .snowflake import Snowflake


class Emoji:
    def __init__(self, client, resp):
        self.id = Snowflake.optional(resp.get("id"))
        self.name = resp["name"]
        self.roles = [client.get(x) for x in resp.get("roles", [])] if client.has_cache else [Snowflake.optional(x) for x in resp.get("roles", [])]
        self.__user = resp.get("user")
        self.user = User.create(client, self.__user) if self.__user else self.__user
        self.require_colons = resp.get("require_colons")
        self.managed = resp.get("managed")
        self.animated = resp.get("animated")
        self.available = resp.get("available")

    def __str__(self):
        return f"<:{self.name}:{self.id}>" if self.id else self.name

    def __int__(self):
        return int(self.id)



import typing
from .model.snowflake import Snowflake


class CacheContainer:
    def __init__(self, default_expiration_time=None):
        self.default_expiration_time = default_expiration_time
        self.__cache_dict: typing.Dict[str, typing.Union[dict, CacheStorage]] = {"guild_cache": {}}

    def get(self, snowflake_id: typing.Union[str, int, Snowflake], storage_type: str = None, *, ignore_expiration=True):
        if storage_type:
            return self.get_storage(storage_type).get(snowflake_id, ignore_expiration=ignore_expiration)
        for x in self.__cache_dict.values():
            if isinstance(x, dict):
                continue
            res = x.get(snowflake_id, ignore_expiration=ignore_expiration)
            if res:
                return res

    def get_storage(self, storage_type: str):
        if storage_type == "guild_cache":
            return self.__cache_dict["guild_cache"]
        if storage_type not in self.__cache_dict:
            self.__cache_dict[storage_type] = CacheStorage()
        return self.__cache_dict.get(storage_type)

    def get_guild_container(self, guild_id: typing.Union[str, int, Snowflake]):
        guild_caches = self.get_storage("guild_cache")
        guild_id = Snowflake.ensure_snowflake(guild_id)
        if guild_id not in guild_caches:
            guild_caches[guild_id] = GuildCacheContainer(default_expiration_time=self.default_expiration_time)
        return guild_caches[guild_id]

    def add(self, snowflake_id: typing.Union[str, int, Snowflake], obj_type: str, obj, expire_at=None):
        if not expire_at:
            expire_at = self.default_expiration_time

        if obj_type not in self.__cache_dict:
            self.__cache_dict[obj_type] = CacheStorage()
        self.__cache_dict[obj_type].add(Snowflake.ensure_snowflake(snowflake_id), obj, expire_at)

    def remove(self, snowflake_id: typing.Union[str, int, Snowflake], obj_type: str):
        if obj_type in self.__cache_dict:
            self.__cache_dict[obj_type].remove(snowflake_id)
        if "guild_cache" in self.__cache_dict:
            for x in self.__cache_dict["guild_cache"].values():
                x.remove(snowflake_id, obj_type)

    @property
    def available_cache_types(self):
        return self.__cache_dict.keys()

    @property
    def size(self):
        ret = 0
        for k, v in self.__cache_dict.items():
            if k == "guild_cache":
                for b in v.values():
                    ret += b.size
                continue
            ret += v.size
        return ret


class GuildCacheContainer(CacheContainer):
    def get_guild_container(self, *args, **kwargs):
        raise NotImplementedError


class CacheStorage:
    def __init__(self):
        self.__cache_dict = {}

    def get(self, snowflake_id: typing.Union[str, int, Snowflake], *, ignore_expiration=True):
        res = self.__cache_dict.get(Snowflake.ensure_snowflake(snowflake_id))
        if res:  # TODO: add expiration time check
            return res["value"]

    def add(self, snowflake_id: typing.Union[str, int, Snowflake], obj, expire_at=None):
        snowflake_id = Snowflake.ensure_snowflake(snowflake_id)
        self.__cache_dict[snowflake_id] = {"value": obj, "expire_at": expire_at}

    def remove(self, snowflake_id: typing.Union[str, int, Snowflake]):
        snowflake_id = Snowflake.ensure_snowflake(snowflake_id)
        if snowflake_id in self.__cache_dict:
            del self.__cache_dict[snowflake_id]

    @property
    def size(self):
        return len(self.__cache_dict)



from .utils import ensure_coro
from .model.event import *


class EventHandler:
    def __init__(self, client):
        self.events = {}
        self.client = client

    def add(self, event, func):
        if event not in self.events:
            self.events[event] = []

        self.events[event].append(ensure_coro(func))

    def get(self, event) -> list:
        return self.events.get(event, [])

    def process_response(self, name, resp):
        model_dict = {
            "READY": Ready,
            "CHANNEL_CREATE": ChannelCreate,
            "CHANNEL_UPDATE": ChannelUpdate,
            "CHANNEL_DELETE": ChannelDelete,
            "CHANNEL_PINS_UPDATE": ChannelPinsUpdate,
            "GUILD_CREATE": GuildCreate,
            "GUILD_UPDATE": GuildUpdate,
            "GUILD_DELETE": GuildDelete,
            "GUILD_BAN_ADD": GuildBanAdd,
            "GUILD_BAN_REMOVE": GuildBanRemove,
            "GUILD_EMOJIS_UPDATE": GuildEmojisUpdate,
            "GUILD_INTEGRATIONS_UPDATE": GuildIntegrationsUpdate,
            "GUILD_MEMBER_ADD": GuildMemberAdd,
            "GUILD_MEMBER_REMOVE": GuildBanRemove,
            "GUILD_MEMBER_UPDATE": GuildMemberUpdate,
            "GUILD_ROLE_CREATE": GuildRoleCreate,
            "GUILD_ROLE_UPDATE": GuildRoleUpdate,
            "GUILD_ROLE_DELETE": GuildRoleDelete,
            "INTERACTION_CREATE": InteractionCreate,
            "INVITE_CREATE": InviteCreate,
            "INVITE_DELETE": InviteDelete,
            "MESSAGE_CREATE": MessageCreate,
            "MESSAGE_UPDATE": MessageUpdate,
            "MESSAGE_DELETE": MessageDelete,
            "MESSAGE_DELETE_BULK": MessageDeleteBulk,
            "MESSAGE_REACTION_ADD": MessageReactionAdd,
            "MESSAGE_REACTION_REMOVE": MessageReactionRemove,
            "MESSAGE_REACTION_REMOVE_ALL": MessageReactionRemoveAll,
            "MESSAGE_REACTION_REMOVE_EMOJI": MessageReactionRemoveEmoji,
            "PRESENCE_UPDATE": PresenceUpdate,
            "TYPING_START": TypingStart,
            "USER_UPDATE": UserUpdate,
            "VOICE_STATE_UPDATE": VoiceStateUpdate
        }
        if name in model_dict:
            ret = model_dict[name].create(self.client, resp)
        else:
            ret = resp
        return ret

    def dispatch_from_raw(self, name, resp):
        ret = self.process_response(name, resp)
        if hasattr(ret, "_dont_dispatch") and ret._dont_dispatch:
            return
        self.client.dispatch(name, ret)



import typing
from .slashcommands import ApplicationCommandInteractionDataResolved
from .components import Component, ComponentTypes
from ..channel import Embed, AllowedMentions
from ..guild import GuildMember
from ..snowflake import Snowflake
from ..user import User
from ...base.model import TypeBase, FlagBase


class Interaction:
    def __init__(self, client, resp):
        self.client = client
        self.id = Snowflake(resp["id"])
        self.application_id = Snowflake(resp["application_id"])
        self.type = InteractionType(resp["type"])
        self.data = ApplicationCommandInteractionData(client, resp.get("data")) if "data" in resp else resp.get("data")
        self.guild_id = Snowflake(resp.get("guild_id"))
        self.channel_id = Snowflake(resp.get("channel_id"))
        self.__member = resp.get("member")
        self.member = GuildMember.create(client, self.__member, guild_id=self.guild_id) if self.__member else None
        self.__user = resp.get("user")
        self.user = User.create(client, self.__user) if self.__user else None
        self.token = resp["token"]
        self.version = resp["version"]

    def __int__(self):
        return int(self.id)

    def create_response(self, interaction_response):
        return self.client.create_interaction_response(self, interaction_response)

    def create_followup_message(self, **kwargs):
        return self.client.create_followup_message(self, **kwargs)

    @classmethod
    def create(cls, client, resp):
        return cls(client, resp)


class InteractionType(TypeBase):
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3


class ApplicationCommandInteractionData:
    def __init__(self, client, resp: dict):
        self.id = Snowflake.optional(resp.get("id"))
        self.name = resp.get("name")
        self.__resolved = resp.get("resolved")
        self.resolved = ApplicationCommandInteractionDataResolved(client, resp.get("resolved")) if self.__resolved else self.__resolved
        self.options = resp.get("options")
        self.custom_id = resp.get("custom_id")
        self.component_type = ComponentTypes(resp.get("component_type")) if resp.get("component_type") else None
        self.values = resp.get("values", [])


class InteractionCallbackType(TypeBase):
    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7


class InteractionApplicationCommandCallbackDataFlags(FlagBase):
    EPHEMERAL = 1 << 6


class InteractionApplicationCommandCallbackData:
    def __init__(self, *,
                 tts: bool = None,
                 content: str = None,
                 embeds: typing.List[Embed] = None,
                 allowed_mentions: AllowedMentions = None,
                 flags: typing.Union[InteractionApplicationCommandCallbackDataFlags, int] = None,
                 components: typing.List[typing.Union[dict, Component]] = None):
        self.tts = tts
        self.content = content
        self.embeds = embeds
        self.allowed_mentions = allowed_mentions
        self.flags = InteractionApplicationCommandCallbackDataFlags.from_value(flags) if isinstance(flags, int) else flags
        self.components = components

    def to_dict(self):
        ret = {}
        if self.tts is not None:
            ret["tts"] = self.tts
        if self.content is not None:
            ret["content"] = self.content
        if self.embeds is not None:
            ret["embeds"] = [x.to_dict() for x in self.embeds]
        if self.allowed_mentions is not None:
            ret["allowed_mentions"] = self.allowed_mentions.to_dict()
        if self.flags is not None:
            ret["flags"] = int(self.flags)
        if self.components is not None:
            ret["components"] = [x if isinstance(x, dict) else x.to_dict() for x in self.components]
        return ret


class InteractionResponse:
    def __init__(self, callback_type: typing.Union[int, InteractionCallbackType], data: typing.Union[dict, InteractionApplicationCommandCallbackData]):
        self.type = callback_type
        self.data = data

    def to_dict(self):
        return {"type": int(self.type), "data": self.data if isinstance(self.data, dict) else self.data.to_dict()}


class MessageInteraction:
    def __init__(self, client, resp):
        self.id = Snowflake(resp["id"])
        self.type = InteractionType(resp["type"])
        self.name = resp["name"]
        self.user = User.create(client, resp["user"])



from ..base.model import DiscordObjectBase, FlagBase, TypeBase
from ..utils import cdn_url


class User(DiscordObjectBase):
    def __init__(self, client, resp):
        super().__init__(client, resp)
        self._cache_type = "user"
        self.username = resp.get("username")
        self.discriminator = resp.get("discriminator")
        self.avatar = resp.get("avatar")
        self.bot = resp.get("bot", False)
        self.system = resp.get("system", False)
        self.mfa_enabled = resp.get("mfa_enabled", False)
        self.locale = resp.get("locale")
        self.verified = resp.get("verified", False)
        self.email = resp.get("email")
        self.flags = UserFlags.from_value(resp.get("flags", 0))
        self.premium_type = PremiumTypes(resp.get("premium_type", 0))
        self.public_flags = UserFlags.from_value(resp.get("public_flags", 0))

        self.voice_state = self.raw.get("voice_state")  # Filled later.

    def __str__(self):
        return f"{self.username}#{self.discriminator}"

    @property
    def mention(self):
        return f"<@{self.id}>"

    def avatar_url(self, *, extension="webp", size=1024):
        if self.avatar:
            return cdn_url("avatars/{user_id}", image_hash=self.avatar, extension=extension, size=size, user_id=self.id)
        else:
            return cdn_url("embed/avatars", image_hash=self.discriminator % 5, extension=extension)

    def set_voice_state(self, voice_state):
        self.voice_state = voice_state
        self.raw["voice_state"] = voice_state

    def get_voice_state(self):
        return self.voice_state


class UserFlags(FlagBase):
    NONE = 0
    DISCORD_EMPLOYEE = 1 << 0
    PARTNERED_SERVER_OWNER = 1 << 1
    HYPESQUAD_EVENTS = 1 << 2
    BHG_HUNTER_LEVEL_1 = 1 << 3
    HOUSE_BRAVERY = 1 << 6
    HOUSE_BRILLIANCE = 1 << 7
    HOUSE_BALANCE = 1 << 8
    EARLY_SUPPORTER = 1 << 9
    TEAM_USER = 1 << 10
    BHG_HUNTER_LEVEL_2 = 1 << 14
    VERIFIED_BOT = 1 << 16
    EARLY_VERIFIED_BOT_DEVELOPER = 1 << 17


class PremiumTypes(TypeBase):
    NONE = 0
    NITRO_CLASSIC = 1
    NITRO = 2



import copy
from ..model.snowflake import Snowflake


class CopyableObject:
    def copy(self):
        return copy.copy(self)


class EventBase:
    def __init__(self, client, resp: dict):
        self.raw = resp
        self.client = client
        self._dont_dispatch = False

    @classmethod
    def create(cls, client, resp: dict):
        return cls(client, resp)


class DiscordObjectBase:
    def __init__(self, client, resp, **kwargs):
        resp.update(kwargs)
        self._cache_type = None
        self.raw = resp
        self.id = Snowflake(resp["id"])
        self.client = client

    def __int__(self):
        return int(self.id)

    @classmethod
    def create(cls, client, resp, **kwargs):
        maybe_exist = client.has_cache and client.cache.get(resp["id"])
        if maybe_exist:
            orig = maybe_exist.raw
            for k, v in resp.items():
                if orig.get(k) != v:
                    orig[k] = v
            maybe_exist.__init__(client, orig, **kwargs)
            return maybe_exist
        else:
            ret = cls(client, resp, **kwargs)
            if client.has_cache:
                client.cache.add(ret.id, ret._cache_type, ret)
                if hasattr(ret, "guild_id") and ret.guild_id:
                    client.cache.get_guild_container(ret.guild_id).add(ret.id, ret._cache_type, ret)
            return ret


class FlagBase:
    def __init__(self, *args, **kwargs):
        self.values = {x: getattr(self, x) for x in dir(self) if isinstance(getattr(self, x), int)}
        self.value = 0
        for x in args:
            if x.upper() not in self.values:
                raise AttributeError(f"invalid name: `{x}`")
            self.value |= self.values[x.upper()]
        for k, v in kwargs.items():
            if k.upper() not in self.values:
                raise AttributeError(f"invalid name: `{k}`")
            if v:
                self.value |= self.values[k.upper()]

    def __int__(self):
        return self.value

    def __getattr__(self, item):
        return self.has(item)

    def has(self, name: str):
        if name.upper() not in self.values:
            raise AttributeError(f"invalid name: `{name}`")
        return (self.value & self.values[name.upper()]) == self.values[name.upper()]

    def __setattr__(self, key, value):
        orig = key
        key = key.upper()
        if orig in ["value", "values"] or key not in self.values.keys():
            return super().__setattr__(orig, value)
        if not isinstance(value, bool):
            raise TypeError(f"only type `bool` is supported.")
        has_value = self.has(key)
        if value and not has_value:
            self.value |= self.values[key]
        elif not value and has_value:
            self.value &= ~self.values[key]

    def add(self, value):
        return self.__setattr(value, True)

    def remove(self, value):
        return self.__setattr(value, False)

    @classmethod
    def from_value(cls, value: int):
        ret = cls()
        ret.value = value
        return ret


class TypeBase:
    def __init__(self, value):
        self.values = {getattr(self, x): x for x in dir(self) if isinstance(getattr(self, x), int)}
        self.value = value

        if self.value not in self.values:
            raise AttributeError(f"invalid value: `{value}`")

    def __str__(self):
        return self.values[self.value]

    def __int__(self):
        return self.value

    def __getattr__(self, item):
        return self.is_type(item)

    def is_type(self, name: str):
        values = {y: x for x, y in self.values.items()}
        if name.upper() not in values:
            raise AttributeError(f"invalid name: `{name}`")
        return self.value == values[name.upper()]

    @classmethod
    def to_string(cls, value):
        values = {getattr(cls, x): x for x in dir(cls) if isinstance(getattr(cls, x), int)}
        return values.get(value)



import typing

from ..emoji import Emoji
from ..snowflake import Snowflake
from ...base.model import TypeBase


class Component:
    def __init__(self, component_type: typing.Union[int, "ComponentTypes"]):
        self.type = ComponentTypes(component_type) if isinstance(component_type, int) else component_type

    def to_dict(self):
        return {"type": self.type.value}

    @staticmethod
    def auto_detect(resp):
        if resp["type"] == ComponentTypes.ACTION_ROW:
            return ActionRow.create(resp)
        elif resp["type"] == ComponentTypes.BUTTON:
            return Button.create(resp)
        elif resp["type"] == ComponentTypes.SELECT_MENU:
            return SelectMenu.create(resp)
        else:
            raise NotImplementedError


class ComponentTypes(TypeBase):
    ACTION_ROW = 1
    BUTTON = 2
    SELECT_MENU = 3


class ActionRow(Component):
    def __init__(self, *components: typing.Union[Component, dict]):
        super().__init__(ComponentTypes.ACTION_ROW)
        self.components = [Component.auto_detect(x) if isinstance(x, dict) else x for x in components or []]

    def to_dict(self):
        return {"type": self.type.value, "components": [x.to_dict() for x in self.components]}

    @classmethod
    def create(cls, resp):
        return cls(*[Component.auto_detect(x) for x in resp["components"]])


class Button(Component):
    def __init__(self,
                 *,
                 style: typing.Union["ButtonStyles", int],
                 label: str = None,
                 emoji: typing.Union[Emoji, dict, "PartialEmoji"] = None,
                 custom_id: str = None,
                 url: str = None,
                 disabled: bool = False,
                 **_):  # Dummy.
        super().__init__(ComponentTypes.BUTTON)
        self.style = ButtonStyles(style) if isinstance(style, int) else style
        self.label = label
        self.emoji = PartialEmoji(emoji) if isinstance(emoji, dict) else PartialEmoji.from_full_emoji(emoji) if isinstance(emoji, Emoji) else emoji
        self.custom_id = custom_id
        self.url = url
        self.disabled = disabled

    def to_dict(self):
        ret = {"type": self.type.value}
        if self.style is not None:
            ret["style"] = self.style.value
        if self.label is not None:
            ret["label"] = self.label
        if self.emoji is not None:
            ret["emoji"] = {"name": self.emoji.name, "id": self.emoji.id, "animated": self.emoji.animated or False}  # We only need name, id, and animated
        if self.custom_id is not None:
            ret["custom_id"] = self.custom_id
        if self.url is not None:
            ret["url"] = self.url
        if self.disabled is not None:
            ret["disabled"] = self.disabled
        return ret

    @classmethod
    def create(cls, resp):
        return cls(**resp)


class ButtonStyles(TypeBase):
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


class SelectMenu(Component):
    def __init__(self,
                 *,
                 custom_id: str,
                 options: typing.List[typing.Union["SelectOptions", dict]],
                 placeholder: str = None,
                 min_values: int = None,
                 max_values: int = None,
                 **_):  # Dummy.
        super().__init__(ComponentTypes.SELECT_MENU)
        self.custom_id = custom_id
        self.options = [SelectOptions.create(x) if isinstance(x, dict) else x for x in options]
        self.placeholder = placeholder
        self.min_values = min_values
        self.max_values = max_values

    def to_dict(self):
        ret = {"type": self.type.value}
        if self.custom_id is not None:
            ret["custom_id"] = self.custom_id
        if self.options is not None:
            ret["options"] = [x.to_dict() for x in self.options]
        if self.placeholder is not None:
            ret["placeholder"] = self.placeholder
        if self.min_values is not None:
            ret["min_values"] = self.min_values
        if self.max_values is not None:
            ret["max_values"] = self.max_values
        return ret

    @classmethod
    def create(cls, resp):
        return cls(**resp)


class SelectOptions:
    def __init__(self,
                 *,
                 label: str,
                 value: str,
                 description: str = None,
                 emoji: typing.Union["PartialEmoji", Emoji, dict] = None,
                 default: bool = None):
        self.label = label
        self.value = value
        self.description = description
        self.emoji = PartialEmoji(emoji) if isinstance(emoji, dict) else PartialEmoji.from_full_emoji(emoji) if isinstance(emoji, Emoji) else emoji
        self.default = default

    def to_dict(self):
        ret = {"label": self.label, "value": self.value}
        if self.description is not None:
            ret["description"] = self.description
        if self.emoji is not None:
            ret["emoji"] = {"name": self.emoji.name, "id": self.emoji.id, "animated": self.emoji.animated or False}  # We only need name, id, and animated
        if self.default is not None:
            ret["default"] = self.default
        return ret

    @classmethod
    def create(cls, resp):
        return cls(**resp)


class PartialEmoji:
    def __init__(self, resp):
        self.name = resp["name"]
        self.id = Snowflake.optional(resp.get("id"))
        self.animated = resp.get("animated")

    @classmethod
    def from_full_emoji(cls, emoji: Emoji):
        return cls({"name": emoji.name, "id": int(emoji.id), "animated": emoji.animated})



from .snowflake import Snowflake
from ..base.model import DiscordObjectBase, FlagBase


class PermissionFlags(FlagBase):
    CREATE_INSTANT_INVITE = 0x0000000001
    KICK_MEMBERS = 0x0000000002
    BAN_MEMBERS = 0x0000000004
    ADMINISTRATOR = 0x0000000008
    MANAGE_CHANNELS = 0x0000000010
    MANAGE_GUILD = 0x0000000020
    ADD_REACTIONS = 0x0000000040
    VIEW_AUDIT_LOG = 0x0000000080
    PRIORITY_SPEAKER = 0x0000000100
    STREAM = 0x0000000200
    VIEW_CHANNEL = 0x0000000400
    SEND_MESSAGES = 0x0000000800
    SEND_TTS_MESSAGES = 0x0000001000
    MANAGE_MESSAGES = 0x0000002000
    EMBED_LINKS = 0x0000004000
    ATTACH_FILES = 0x0000008000
    READ_MESSAGE_HISTORY = 0x0000010000
    MENTION_EVERYONE = 0x0000020000
    USE_EXTERNAL_EMOJIS = 0x0000040000
    VIEW_GUILD_INSIGHTS = 0x0000080000
    CONNECT = 0x0000100000
    SPEAK = 0x0000200000
    MUTE_MEMBERS = 0x0000400000
    DEAFEN_MEMBERS = 0x0000800000
    MOVE_MEMBERS = 0x0001000000
    USE_VAD = 0x0002000000
    CHANGE_NICKNAME = 0x0004000000
    MANAGE_NICKNAMES = 0x0008000000
    MANAGE_ROLES = 0x0010000000
    MANAGE_WEBHOOKS = 0x0020000000
    MANAGE_EMOJIS = 0x0040000000
    USE_SLASH_COMMANDS = 0x0080000000
    REQUEST_TO_SPEAK = 0x0100000000
    MANAGE_THREADS = 0x0400000000
    USE_PUBLIC_THREADS = 0x0800000000
    USE_PRIVATE_THREADS = 0x1000000000


class Role(DiscordObjectBase):
    def __init__(self, client, resp, *, guild_id=None):
        super().__init__(client, resp)
        self.guild_id = Snowflake.optional(guild_id)  # This isn't actually in payload, but role is always created at the guild, so why not?
        self._cache_type = "role"
        self.name = resp["name"]
        self.color = resp["color"]
        self.hoist = resp["hoist"]
        self.position = resp["position"]
        self.permissions = PermissionFlags.from_value(int(resp["permissions"]))
        self.managed = resp["managed"]
        self.mentionable = resp["mentionable"]
        self.tags = RoleTags.optional(resp.get("tags"))

    @property
    def guild(self):
        if self.client.has_cache:
            return self.client.get(self.guild_id, "guild")


class RoleTags:
    def __init__(self, resp):
        self.bot_id = Snowflake.optional(resp.get("bot_id"))
        self.integration_id = Snowflake.optional(resp.get("integration_id"))
        self.premium_subscriber = resp.get("premium_subscriber")

    @classmethod
    def optional(cls, resp):
        if resp:
            return cls(resp)



import sys
import typing
import asyncio
import traceback
from . import utils
from .api import APIClient
from .http.async_http import AsyncHTTPRequest
from .ws.websocket import WebSocketClient
from .cache import CacheContainer
from .exception import WebsocketClosed
from .handler import EventHandler
from .model import Intents, AllowedMentions, Snowflake, Application, Activity, Guild, Channel


class Client(APIClient):
    """
    Client with async request handler and websocket support.

    :param token: Token of the client.
    :param intents: Intents to use. Default everything except privileged.
    :param default_allowed_mentions: Default :class:`.model.channel.AllowedMentions` object to use. Default None.
    :param loop: asyncio Event loop object to use. Default automatic.
    :param application_id: Application ID if needed.

    :ivar http: HTTP request client.
    :ivar default_allowed_mentions: Default :class:`.model.channel.AllowedMentions` object of the client.
    :ivar loop: asyncio Event loop object of the client.
    :ivar cache: :class:`.cache.Intents` of the client.
    :ivar intents: :class:`.model.gateway.Intents` of the client.
    :ivar ws: Websocket client of the client.
    :ivar events: :class:`.handler.EventHandler` of the client.
    :ivar application: :class.model.gateway.Application: of the client.
    :ivar application_id: ID of the application. Can be ``None`` if Ready event is not called, and if it is, you must pass parameter application_id for all methods that has it.
    """

    def __init__(self,
                 token: str, *,
                 intents: Intents = Intents.no_privileged(),
                 default_allowed_mentions: AllowedMentions = None,
                 loop=None,
                 cache: bool = True,
                 application_id: typing.Union[str, int, Snowflake] = None):
        self.loop = loop or asyncio.get_event_loop()
        super().__init__(token, base=AsyncHTTPRequest, default_allowed_mentions=default_allowed_mentions, loop=loop)
        self.token = token
        self.__use_cache = cache
        self.cache = CacheContainer() if self.__use_cache else None  # The reason only supporting caching with WS is to ensure up-to-date object.
        self.__ws_class = WebSocketClient
        self.intents = intents
        self.ws: typing.Union[None, WebSocketClient] = None
        self.events = EventHandler(self)
        self.application: typing.Union[None, Application] = None
        self.__wait_futures = {}
        self.application_id = Snowflake.ensure_snowflake(application_id)

        # Internal events dispatch
        self.events.add("READY", self.__ready)
        self.events.add("VOICE_STATE_UPDATE", self.__voice_state_update)

    def __ready(self, ready):
        self.application_id = Snowflake(ready.application["id"])

    def __voice_state_update(self, voice_state):
        if self.has_cache:
            user = self.get(voice_state.user_id)
            if user:
                user.set_voice_state(voice_state)

    def on_(self, name: str = None, meth=None):
        """
        Adds new event listener. This can be used as decorator or function.

        :param name: Name of the event. Case-insensitive. Default name of the function.
        :param meth: Method or Coroutine, if you don't want to use as decorator.
        """
        def wrap(func=None):
            func = func or meth
            self.events.add(name.upper() if name else func.__name__.upper().lstrip("ON_"), func)
            return func
        return wrap if meth is None else wrap()

    @property
    def on(self):
        """Alias of :meth:`.on_`"""
        return self.on_

    def wait(self, event_name: str, timeout: float = None, check: typing.Callable[[typing.Any], bool] = None):
        """
        Waits for the event dispatch.

        :param event_name: Name of the event. Case insensitive.
        :param timeout: Timeout time in second. If not passed, it will wait forever.
        :param check: Check function of the event. If passed, it will wait until the event result passes the check.
        :return: Payload of the event.
        :raises: TimeoutError - Timeout occurred.
        :raises: WebsocketClosed - Websocket is closed, so waiting is canceled.
        """
        async def wrap():
            while not self.websocket_closed:
                future = asyncio.Future()
                if event_name.upper() not in self.__wait_futures:
                    self.__wait_futures[event_name.upper()] = []
                self.__wait_futures[event_name.upper()].append(future)
                res = await asyncio.wait_for(future, timeout=None, loop=self.loop)
                ret = res if len(res) > 1 else res[0]
                if check and check(*res):
                    return ret
                elif not check:
                    return ret
            raise WebsocketClosed
        return asyncio.wait_for(wrap(), timeout=timeout, loop=self.loop)

    def dispatch(self, name, *args):
        """
        Dispatches new event.

        :param name: Name of the event.
        :param args: Arguments of the event.
        """
        [self.loop.create_task(utils.safe_call(x(*args))) for x in self.events.get(name.upper())]
        # [self.__wait_futures[name.upper()].pop(x).set_result(args) for x in range(len(self.__wait_futures.get(name.upper(), [])))]
        for x in range(len(self.__wait_futures.get(name.upper(), []))):
            fut: asyncio.Future = self.__wait_futures[name.upper()].pop(x)
            if not fut.cancelled():
                fut.set_result(args)

    @property
    def get(self):
        """Alias of ``.cache.get``."""
        if self.has_cache:
            return self.cache.get

    async def start(self, reconnect_on_unknown_disconnect: bool = False):
        """
        Starts websocket connection and clears every connections after stopping due to error or KeyboardInterrupt.

        .. warning::
            You can call this only once.
        """
        self.ws = await self.__ws_class.connect(self.http, self.intents, self.events, reconnect_on_unknown_disconnect)
        try:
            await self.ws.run()
        except KeyboardInterrupt:
            print("Detected KeyboardInterrupt, exiting...", file=sys.stderr)
        except Exception as ex:
            print("Unexpected exception occurred, exiting...", file=sys.stderr)
            traceback.print_exc()
        finally:
            await self.ws.close()
            await self.http.close()

    async def close(self):
        """Clears all connections and closes session."""
        await self.ws.close()
        await self.http.close()

    def update_presence(self, *, since: int = None, activities: typing.List[typing.Union[Activity, dict]], status: str = "online", afk: bool = False):
        """
        Updates the bot presence.

        All parameters must be passed as keyword.

        :param since: Time as millisecond when the bot was idle since.
        :param activities: List of activities.
        :param status: Status of the bot.
        :param afk: Whether the bot is AFK.
        """
        activities = [x.to_dict() if not isinstance(x, dict) else x for x in activities]
        return self.ws.update_presence(since, activities, status, afk)

    def update_voice_state(self,
                           guild: typing.Union[int, str, Snowflake, Guild],
                           channel: typing.Union[int, str, Snowflake, Channel] = None,
                           self_mute: bool = False,
                           self_deaf: bool = False):
        """
        Changes the voice state of the bot in guild. (Connecting/Disconnecting from the guild, etc...)

        :param guild: Guild to change presence.
        :param channel: Voice channel to connect. Pass nothing or None to disconnect from the channel.
        :param self_mute: Whether the bot is self-muted.
        :param self_deaf: Whether the bot is self-deaf.
        """
        return self.ws.update_voice_state(str(int(guild)), str(int(channel)) if channel else None, self_mute, self_deaf)

    @property
    def has_cache(self):
        return self.__use_cache

    @property
    def websocket_closed(self):
        """
        Whether the bot is disconnected from the Discord websocket.
        :return: bool
        """
        if self.ws:
            return self.ws.closed
        return True

    def __setattr__(self, key, value):
        if not key.lower().startswith("on_") or key.lower() in ["on", "on_"]:
            return super().__setattr__(key, value)
        event_name = key.lower().lstrip("on_")
        return self.on_(event_name, value)

    def __getattr__(self, item):
        if not item.lower().startswith("on_") or item.lower() in ["on", "on_"]:
            return super().__getattribute__(item)  # Should raise AttributeError or whatever.
        event_name = item.lower().lstrip("on_")

        def deco(func):
            return self.on_(event_name, func)

        return deco

    def run(self, reconnect_on_unknown_disconnect: bool = False):
        """
        Runs client. Actually this is sync function wrapping :meth:`.start`.

        .. warning::
            This must be placed at the end of the code.
        """
        self.loop.run_until_complete(self.start(reconnect_on_unknown_disconnect))
